# TracIncludeMacro macros

import os.path
import re
import urllib2
from StringIO import StringIO
import copy

from trac.core import *
from trac.env import open_environment
from trac.config import Option, BoolOption, ListOption
from trac.wiki.macros import WikiMacroBase
from trac.wiki.api import parse_args 
from trac.wiki.formatter import system_message
from trac.wiki.model import WikiPage
from trac.mimeview.api import Mimeview, get_mimetype, Context
from trac.perm import IPermissionRequestor, PermissionCache
from genshi.core import escape
from genshi.input import HTMLParser, ParseError
from genshi.filters.html import HTMLSanitizer
from trac.util.text import unicode_quote
from trac.web.href import Href
from genshi.filters.transform import Transformer
from genshi.path import PathSyntaxError
try:
    # BeautifulSoup4
    from bs4 import BeautifulSoup
    # bs4 changed params signature from fromEncoding to from_encoding
    _bs_from_encoding_kwname = 'from_encoding'
except ImportError:
    # fall back to BeautifulSoup3
    try:
        from BeautifulSoup import BeautifulSoup
        # bs3 expects fromEncoding instead of from_encoding
        # we could also inherit from BeautifulSoup and adapt parameter naming
        # in a custom __init__ method but I guess it's simpler this way 
        _bs_from_encoding_kwname = 'fromEncoding'
    except ImportError:
        BeautifulSoup = None


__all__ = ['IncludeMacro']


# URL regexp adapted from http://www.ietf.org/rfc/rfc3986.txt Appendix B
# the re named groups represent the five url components described there,
# the additional pqf named group is a shortcut for <path>?<query>#<fragment>
url_re = re.compile(
    '^((?P<scheme>[^:/?#]+):)?(//(?P<authority>[^/?#]*))?'
    '(?P<pqf>(?P<path>[^?#]*)(\?(?P<query>[^#]*))?(#(?P<fragment>.*))?)').match


# Isn't there a cleaner way to do this?
# Might just as well use a function.
class HijackedRequest(object):
    """Factory class that returns hijacked request objects.
    
    Returns a shallow copy of the req arg with modified req.environ entries
    according to environ (may be given as keyword args).
    Adapts the request attributes base_url, href and abs_href to the updated
    environ.
    """
    def __new__(cls, req, **environ):
        hijacked_req = copy.copy(req)
        hijacked_req.environ.update(environ)
        hijacked_req.base_url = hijacked_req.environ.get('trac.base_url')
        # code adapted from Request.__init__
        if not hijacked_req.base_url:
            hijacked_req.base_url = hijacked_req._reconstruct_url()
        hijacked_req.href = Href(hijacked_req.base_path)
        hijacked_req.abs_href = Href(hijacked_req.base_url)
        return hijacked_req


def make_absolute_links(url):
    """Return genshi relative-to-absolute link attribute manipulator callable.
     
    The callable can be used in genshi attribute transformations to convert
    attributes containing URL links to absolute links by using the `url` 
    authority and path. Links may be either relative or "absolute server-local"
    (i.e. starting with '/').
       
    Usage, e.g. for href attributes:
    out = out | Transformer('//*[@href]').attr('href', abs_links_transformer)
    """
     
    urlparts = url_re(url).groupdict('')
    def abs_links_transformer(name, event):
        attrs = event[1][1]
        link = attrs.get(name)
        if link:
            linkparts = url_re(link).groupdict('')
            if not linkparts['authority']:
                # server-local path 
                link_pqf = linkparts['pqf']
                if link_pqf.startswith('/'):
                    # absolute server-local path
                    pqf = link_pqf
                else:
                    # relative server-local path
                    # use path part of url
                    pqf = urlparts['path'].rstrip('/') + '/' + link_pqf
                link = linkparts['scheme'] or urlparts['scheme'] + '://' + \
                    urlparts['authority'] + pqf
        # fallback: link contains an absolute path anyway, or element did not
        # previously carry the attribute <name>
        return link
    return abs_links_transformer


class IncludeMacro(WikiMacroBase):
    """A macro to include other resources in wiki pages.
   
    Currently these source prefixes are supported for inclusion:
     * HTTP - http: and https: 
     * FTP - ftp:
     * Wiki pages - wiki:
     * Repository files - source:

    The default prefix is `wiki` if only a source path is given.
    Intertrac prefixes may be used for including resources from other known
    Trac instances.

    Examples:
    {{{
    [[Include(SomeWikiPage)]]             # wiki page SomeWikiPage
    [[Include(wiki:SomeWikiPage]]         # wiki page SomeWikiPage
    [[Include(i_other:wiki:WikiStart)]]   # intertrac wiki page WikiStart 
    [[Include(http://trac.edgewall.org)]] # http resource
    [[Include(http://bogus.encodings.org, from_encoding='utf-8')]] # fix enc.
    [[Include(source:trunk/version.txt)]] # repo source file 
    [[Include(http://site.com, user=user, pass=pass))]] # auth. http site 
    }}}

    Optionally uses BeautifulSoup to fix broken included (x)html, if installed
    and activated in the macro configuration. 
    
    An optional second argument sets the output MIME type, though in most
    cases the default will be correct.
    The three remote sources (http, https, and ftp) require INCLUDE_URL to be
    rendered. For remote sources, relative and server-local absolute URLs are
    converted to full absolute URLs (currently for */@href, img/@src and 
    form/@action link attributes).
    The optional `xpath` keyword argument can be used to define a genshi 
    transformation on the output html to only select parts of it. Only elements
    selected by this xpath will be in the included result (genshi
    transformation: `Transformer(xpath).invert().remove()`).
    If the included html doesn't get properly decoded its source encoding may
    be given as a hint to BeautifulSoup by using keyword arg `from_encoding`.
    The optional keyword arguments `user` and `pass` can be used for inclusion
    of authenticated sites.
 
    Arguments:
     * source: Resource to include
     * dest_format: Output MIME type (optional)
  
    Keyword arguments:
     * xpath: Only process (x)html elements selected by this genshi Transformer
       xpath expression (optional)
     * from_encoding: set source encoding for external html as a hint for
       BeautifulSoup encoding detection (optional; only applied if beautify_html
       option is set)
     * user: Username for authenticated http site inclusion (optional)
     * pass: Password for authenticated http site inclusion (optional)
    """
    
    implements(IPermissionRequestor)

    # includemacro configuration options
    beautify_html = BoolOption('includemacro', 'beautify', default=False,
                                doc='apply BeautifulSoup (if installed)')
    remote_itrac_filter = Option(
        'includemacro', 'remote_itrac_filter',
         default='body//div[@id="content"]"',
         doc='only process (x)html elements selected by this genshi'
         ' Transformer xpath expression')
    allow_url_re = Option(
        'includemacro', 'allow_url_re', default=None,
        doc='regular expression, generally allow inclusion of matching URLs')
    # Without allowing html & body, no (x)html documents will ever be rendered if
    # render_unsafe_content is off, only (x)html snippets. Does it hurt to 
    # allow body to be parsed? I don't think so, as already non xhtml-valid inclusion 
    # may render due to the parser not yielding events in document order.
    # All offensive stuff inside body will be filtered out nonetheless.
    safe_tags = ListOption(
        'includemacro', 'safe_tags',
        default=HTMLSanitizer.SAFE_TAGS|set(['html', 'body']) ,
        doc='HTMLSanitizer safe_tags configuration (default: '
        'HTMLSanitizer.SAFE_TAGS|set(["html", "body"])')
    safe_attrs = ListOption(
        'includemacro', 'safe_attrs',
        default=HTMLSanitizer.SAFE_ATTRS,
        doc='HTMLSanitizer safe_attrs configuration (default: '
        'HTMLSanitizer.SAFE_ATTRS)')
    
    # Default output formats for sources that need them
    default_formats = {
        'wiki': 'text/x-trac-wiki',
    }

    # regular expression to pull out [prefix:]source_format:source_obj from 
    # include source argument, e.g.
    # >>> include_re("i_other:wiki:MyPage").groupdict()
    # {'prefix': 'i_other', 'source_obj': 'MyPage', 'source_format': 'wiki'}
    # >>> include_re("http://localhost:8080/foo/bar?x=3").groupdict()
    # {'prefix': None, 'source_obj': '//localhost:8080/foo/bar?x=3', 
    #  'source_format': 'http'}
    # >>> include_re("http://www.python.org").groupdict()
    # {'prefix': None, 'source_obj': '//www.python.org', 
    #  'source_format': 'http'}
    include_re = re.compile(
        '\A(((?P<prefix>[^:]+):)?'
        '(?P<source_format>[^:/]+):)?'
        '(?P<source_obj>[^:].+)\Z'
        ).match

    beautifier = BeautifulSoup
 
    def __init__(self):
        self.sanitizer = HTMLSanitizer(
            safe_tags=self.safe_tags, safe_attrs=self.safe_attrs) 

    # IWikiMacroProvider methods
    def expand_macro(self, formatter, name, content):
        args, kw = parse_args(content)
        args = [ arg.strip() for arg in args ]
        if len(args) == 1:
            args.append(None)
        elif len(args) != 2:
            return system_message('Invalid arguments "%s"' % content)
        source, dest_format = args
        xpath = kw.get('xpath') 
        from_encoding = kw.get('from_encoding') 
        username = kw.get('user')
        password = kw.get('pass')

        # pull out include source parts
        match = self.include_re(source)
        if not match:
            return system_message(
                'Unsupported include source format %s' % source)
        sourceinfo = match.groupdict()
        intertrac_prefix = sourceinfo['prefix']
        # assume a wiki page if no source format info given
        source_format = sourceinfo['source_format'] or 'wiki'
        source_obj = sourceinfo['source_obj']
        source = ':'.join([source_format, source_obj])

        included_env = self.env
        included_perm = formatter.perm
        # Hm, trac seems to plan to remove req from formatter in future
        # releases. Where to get base_url information from, then?
        req = formatter.req

        # force absolute links for wiki sources whose context uses abs_href as
        #  href
        absurls = (req.abs_href() == formatter.context.href())

        intertrac_is_local = False
        included_req = req
        abs_links_transformer = None
 
        if intertrac_prefix:
            # map intertrac link to actual http: link
            intertrac = self.env.config['intertrac']
            intertrac_name = intertrac.get(intertrac_prefix, intertrac_prefix)
            intertrac_url = intertrac.get(intertrac_name + '.url')
            # code taken from formatter Formatter._make_intertrac_link(),
            # unfortunately there seems no way to retrieve this info from trac
            # itself
            if not intertrac_url and intertrac_prefix == 'trac':
                intertrac_url = 'http://trac.edgewall.org'
            self.log.debug("intertrac_name=%s, URL=%s"
                           % (intertrac_name, intertrac_url))
            if intertrac_url:
                if intertrac_url.startswith('/'):
                    intertrac_is_local = True
                    url = req.base_url.replace(req.base_path, intertrac_url)
                    projects_dir = os.path.dirname(self.env.path)
                    intertrac_env_name = os.path.basename(intertrac_url)
                    included_env = open_environment(
                        os.path.join(projects_dir, intertrac_env_name))
                    included_perm = PermissionCache(included_env, req.authname)
                    # hijack the request; how *should* this be done?
                    included_req = HijackedRequest(req, 
                        SCRIPT_NAME=intertrac_url,
                        PATH_INFO='/%s/%s' % (source_format, source_obj))
                else:
                    url = intertrac_url
                    # use url schema as source_format for remote intertrac 
                    # includes
                    source_format = url.split(':', 1)[0]

                source = '%s/intertrac/%s' % (url, unicode_quote(source))

        # Apply a default format if needed
        if dest_format is None:
            try:
                # use original source format information to also cover
                # intertrac linking 
                dest_format = self.default_formats[source_format]
            except KeyError:
                pass

        self.log.debug("Include: source_format=%s, dest_format=%s"
                       % (source_format, dest_format))

        if source_format in ('http', 'https', 'ftp'):
            # Since I can't really do recursion checking, and because this
            # could be a source of abuse allow selectively blocking it.
            # RFE: Allow blacklist/whitelist patterns for URLS. <NPK>
            # RFE: Track page edits and prevent unauthorized users from ever 
            # entering a URL include. <NPK>
            if not (formatter.perm.has_permission('INCLUDE_URL') or 
                    self.allow_url_re and re.match(self.allow_url_re, source)):
                self.log.info('IncludeMacro: Blocking attempt by %s to include'
                              ' URL %s on page %s', req.authname, source, 
                              req.path_info)
                return ''
            try:
                self.log.info('IncludeMacro: Opening URL %s', source)
                if username and password:
                    self.log.debug(
                        'IncludeMacro: Opening with user=%s, pass=xxx',
                         username)
                    password_mgr = urllib2.HTTPPasswordMgrWithDefaultRealm() 
 	            password_mgr.add_password(None, source, username, password) 
 	            handler = urllib2.HTTPBasicAuthHandler(password_mgr) 
 	            opener = urllib2.build_opener(handler) 
 	            urlf = opener.open(source)
                else:
                    urlf = urllib2.urlopen(source)
                out = urlf.read()
                # use http headers content encoding information if no explicit
                # from_encoding has been set
                from_encoding = (from_encoding or
                                 urlf.headers.getparam('charset'))
            except urllib2.URLError, e:
                return system_message('Error while retrieving file', str(e))
            ctxt = Context.from_request(included_req)
            # define a genshi attribute transformer callable using the
            # remote server url
            abs_links_transformer = make_absolute_links(source)
        elif source_format == 'wiki':
            # XXX: Check for recursion in page includes. <NPK>
            if not included_perm.has_permission('WIKI_VIEW'):
                return ''
            page = WikiPage(included_env, source_obj)
            if not page.exists:
                return system_message('Wiki page %s does not exist',
                                      source_obj)
            out = page.text
            ctxt = Context.from_request(included_req, 'wiki', source_obj,
                                        absurls=absurls)
        elif source_format == 'source':
            if not included_perm.has_permission('FILE_VIEW'):
                return ''
            repo = included_env.get_repository(req.authname)
            node = repo.get_node(source_obj)
            node_content = node.get_content()
            if not node_content:
                return system_message(
                    'versioncontrol source directory include not supported')
            out = node.get_content().read()
            if dest_format is None:
                dest_format = node.content_type \
                              or get_mimetype(source_obj, out)
            ctxt = Context.from_request(req, 'source', source_obj)
        # RFE: Add ticket: and comment: sources. <NPK>
        # RFE: Add attachment: source. <NPK>
        else:
            return system_message('Unsupported include source %s' 
                                  % source_format)

        # If we have a preview format, use it
        if dest_format:
            out = Mimeview(included_env).render(ctxt, dest_format, out)
        if (self.beautify_html and self.beautifier and 
            isinstance(out, (basestring, buffer))):
            try:
                # BeautifulSoup will use from_encoding if this is applicable and
                # gracefully fall back to its default encoding detection
                # mechanism otherwise
                out = self.beautifier(
                    out, **{_bs_from_encoding_kwname: from_encoding}).prettify()
            except Exception, e:
                # if the beautifier fails try to go on anyway
                self.log.debug('beautifying included html failed:', str(e)) 
        out = HTMLParser(StringIO(out)).parse()
        if intertrac_prefix and not intertrac_is_local \
            and self.remote_itrac_filter:
            try:
                out = out | Transformer(
                    self.remote_itrac_filter).invert().remove() 
            except PathSyntaxError:
                self.log.debug(
                    'IncludeMacro: genshi Transformer PathSyntaxError')

        if abs_links_transformer:
            try:
                out = out | \
                    Transformer('//*[@href]').attr('href', 
                                                   abs_links_transformer) |\
                    Transformer('//img[@src]').attr('src',
                                                      abs_links_transformer) |\
                    Transformer('//form[@action]').attr('action',
                                                        abs_links_transformer)
                    #'//*[@href or @src or @action]') \
                    #.attr('href', abs_links_transformer) \
                    #.attr('src', abs_links_transformer) \
                    #.attr('action', abs_links_transformer)
            except PathSyntaxError:
                self.log.debug(
                    'IncludeMacro: genshi Transformer PathSyntaxError')
        # Escape if needed
        if not self.config.getbool('wiki', 'render_unsafe_content', False):
            try:
                out = out | self.sanitizer
            except ParseError:
                self.log.debug('IncludeMacro: HTMLParser ParseError')
        if xpath:
            out = out | Transformer(xpath).invert().remove()
        out = escape(out)
        return out
 
    # IPermissionRequestor methods
    def get_permission_actions(self):
        yield 'INCLUDE_URL'

